import java.util.*;
import java.io.*;
import java.util.jar.*;
import java.util.zip.*;

import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;

/**
 * This class is checking standard errors in XML files due in TREC's Video Track
 * It permit to launch the parsing an Xml files.
 * The parsing error stop the program while other errors like invalid time format, 
 * stop or start time out of range, etc are of no importance on the rest of checking operations
 *
 * @version $Revision: 0.1 $
 *
 * @author
 * This software was produced by NIST, an agency of the
 * U.S. government, and by statute is not subject to copyright in the
 * United States.  Recipients of this software assume all
 * responsibilities associated with its operation, modification and
 * maintenance.
 *
 */

public class Checker extends DefaultHandler {
    
    public static void main(String argv[]){

        if (argv.length != 1) {
            System.err.println("Usage: java checker xmlfile.xml");
            System.exit(1);
        }
        // Use an instance of ourselves as the SAX event handler

        SAXParserFactory factory = SAXParserFactory.newInstance();

        try {
        	DefaultHandler handler = new Checker(out);
        	DefaultHandler handler2 = new Checker(out);
            out = new OutputStreamWriter(System.out, "UTF8");
            SAXParser saxParser1 = factory.newSAXParser();
            
            //Parsing of the data.xml file....
            saxParser1.parse(new File("data.xml"), handler);
            
            factory = SAXParserFactory.newInstance();

            SAXParser saxParser2 = factory.newSAXParser();
            //Parsing of the provided file...
            saxParser2.parse(new File(System.getProperty("user.dir")+
									System.getProperty("file.separator")+argv[0]), handler2);

            //Simple output of errors
            if(sb.equals("") && gs.equals("")){
            	System.out.println("NO ERROR");
            }
            else{
            	System.out.println("ERRORS");
            	System.out.println(sb);
            	System.out.println(gs);
            }

        } catch (SAXParseException spe) {
           // Error generated by the parser
           System.out.println("\n** Parsing error"
              + ", line " + spe.getLineNumber()
              + ", uri " + spe.getSystemId());
           System.out.println("   " + spe.getMessage() );

           // Use the contained exception, if any
           Exception  x = spe;
           if (spe.getException() != null)
               x = spe.getException();
           x.printStackTrace();

        } catch (SAXException sxe) {
           // Error generated by this application
           // (or a parser-initialization error)
           System.out.println("\n** Error generated by this application or a parser-initialization error");
           /*Exception  x = sxe;
           if (sxe.getException() != null)
               x = sxe.getException();
           x.printStackTrace();*/

        } catch (ParserConfigurationException pce) {
            // Parser with specified options can't be built
            System.out.println("\n** Parser-initialization error");
            //pce.printStackTrace();
        } catch (IOException ioe) {
           // I/O error
           System.out.println("\n** Input/Output error");
           //ioe.printStackTrace();
        }

        System.exit(0);
    }
    

    //===========================================================
    // Attributes
    //===========================================================

    static private Writer  out;
    //videosInfos contains vInfos objects which are created for each line in data.xml
    //concerning a video
    static private Hashtable videosInfos = new Hashtable();
    
    //sb: The string containing errors concerning shot boundary file
    //gs: for general search, but not usefull to distinguish these two case    
    static private String sb = "";
    static private String gs = "";
    
    //currentTransvid contains the name of the current video
	private String currentTransvid = "";

	//vInfos is small class to contains information about videos found in data.xml
    private class vInfos{
    
    	public String fc;
    	public String tc;
    	public String name;
    	public vInfos(String name, String fc, String tc){
    		this.name = name; this.fc = fc; this.tc = tc;	
    	}
    	public vInfos(){}
    }
    
	
	public Checker(Writer out){ this.out = out; }

    //===========================================================
    // SAX DocumentHandler methods
    //===========================================================
    public void setDocumentLocator(Locator l){
        // Save this to resolve relative URIs or to give diagnostics.
      /*try {
          //out.write("LOCATOR");
          out.write("    "+l.getSystemId() );
          out.flush();
        } catch (IOException e) {
            // Ignore errors
        }*/
    }

    public void startDocument() throws SAXException{
        nl();
    }

    public void endDocument() throws SAXException {
        /*emit("END DOCUMENT");
        try {
            nl();
            out.flush();
        } catch (IOException e) {
            throw new SAXException("I/O error", e);
        }*/
    }

    public void startElement(String namespaceURI,
                             String lName, // local name
                             String qName, // qualified name
                             org.xml.sax.Attributes attrs) throws SAXException {
        //indentLevel++;
        String eName = lName; // element name
        if ("".equals(eName)) eName = qName; // namespaceAware = false
        

		//Different parsed elements an their attributes are used to construct vInfos Hashtable
		//and compare submitted data with this information : 
		
		if(eName.equals("rv")){
			vInfos aVid = new vInfos();
			String videoName = null;
			if (attrs != null) {
	        	for (int i = 0; i < attrs.getLength(); i++) {
	
	                String aName = attrs.getLocalName(i); // Attr name 
	                if ("".equals(aName)) aName = attrs.getQName(i);
	                if(aName.equals("name")) aVid.name = attrs.getValue(i).toLowerCase();
	                if(aName.equals("fc")) aVid.fc = attrs.getValue(i);
	                if(aName.equals("tc")) aVid.tc = attrs.getValue(i);
	            }
			}
			videosInfos.put(aVid.name, aVid);
		}
		else if(eName.equals("item"))	{
			String vid = "";
			String start = "";
			String stop = "";
			
			if (attrs != null) {
	        	for (int i = 0; i < attrs.getLength(); i++) {
	
	                String aName = attrs.getLocalName(i); // Attr name 
	                if ("".equals(aName)) aName = attrs.getQName(i);
	                if(aName.equals("src")) vid = attrs.getValue(i).toLowerCase();
					if(aName.equals("start")) start = attrs.getValue(i);
					if(aName.equals("stop")) stop = attrs.getValue(i);
	            }
			}
			if(! videosInfos.containsKey(vid)){
				gs += System.getProperty("line.separator") + vid + " is not a valid videos name";
			}
			long startL = transformTime(start);
			long stopL = transformTime(stop);
			if(startL == -1){
				gs += System.getProperty("line.separator") + start + " is not in valid time format";
			}
			if(stopL == -1){
				gs += System.getProperty("line.separator") + stop + " is not in valid time format";
			}
			
			vInfos infos = (vInfos) videosInfos.get(vid);
			if(infos != null){
				long range = transformTime(infos.tc);
				if(stopL > range || stopL < 0){
					gs += System.getProperty("line.separator") + stop + " is an out of range stop time";
				}
				if(startL > range || startL < 0){
					gs += System.getProperty("line.separator") + start + " is an out of range start time";
				}
			}
			if(stopL <= startL){
				gs += System.getProperty("line.separator") + stop + " stop time should not be inferior or equal to "+ start;
			}
				
		}
		else if(eName.equals("seg"))	{
			String vid = "";
			
			if (attrs != null) {
	        	for (int i = 0; i < attrs.getLength(); i++) {
	
	                String aName = attrs.getLocalName(i); // Attr name
	                if ("".equals(aName)) aName = attrs.getQName(i);
	                if(aName.equals("src")) vid = attrs.getValue(i).toLowerCase();
	            }
			}
			if(! videosInfos.containsKey(vid)){
				gs += System.getProperty("line.separator") + vid + " is not a valid video name";
			}
			currentTransvid = vid;
		}
		else if(eName.equals("trans")){

			int pre = 0;
			int post = 0;
			if (attrs != null) {
            	for (int i = 0; i < attrs.getLength(); i++) {

	                String aName = attrs.getLocalName(i); // Attr name 
	                if ("".equals(aName)) aName = attrs.getQName(i);
	                if(aName.equals("preFNum")){
	                	pre = Integer.parseInt((String) attrs.getValue(i));
	                }
	                if(aName.equals("postFNum")) post = Integer.parseInt(attrs.getValue(i));
	            }
			}
			vInfos infos = (vInfos) videosInfos.get(currentTransvid);
			if(infos != null){
				if(post > Integer.parseInt(infos.fc) || post < 0){
					sb += System.getProperty("line.separator") + post + " is an out of range stop frame number";
				}
				if(pre > Integer.parseInt(infos.fc) || pre < 0){
					sb += System.getProperty("line.separator") + pre + " is an out of range start frame number";
				}
			}
			if(post <= pre){
				sb += System.getProperty("line.separator") + post + " frame number should not be inferior or equal to "+ pre;
			}
	    }
    }
    
    //Just used to transform a time in the defined format to a long
	/*public long transformTime(String time) throws NumberFormatException{

		long retour = 0;
		
		try{
			int h=-1;
			int m=-1;
			int s=-1;
			int ms=-1;
			double hours = 0;
			double minutes = 0;
			double seconds = 0;
			double milliseconds = 0;

			if((h = time.indexOf("h")) != -1)
				hours = Double.parseDouble(time.substring(0, h));
			if(time.indexOf("m") != -1)
				if(m+1 < time.length() && time.charAt(m+1) != 's'){
					m = time.indexOf("m");
					minutes = Double.parseDouble(time.substring(h+1, m));
				}
			if(time.indexOf("s") != -1)
				if(time.charAt(time.indexOf("s")-1) != 'm'){
					s = time.indexOf("s");
					seconds = Double.parseDouble(time.substring(m+1, s));
				}
			if((ms = time.indexOf("ms")) != -1)
				milliseconds = Double.parseDouble(time.substring(s+1, ms));
			
			Double d = new Double(hours*60*60*1000 + minutes*60*1000 + seconds*1000 + milliseconds);
			retour = d.longValue();
		} catch (NumberFormatException ex){
			return -1;			
		}
		if(retour !=0)
			return retour;
		else return -1;
	}*/
	public long transformTime(String time) throws NumberFormatException{

		return extractHours(time)+extractMinutes(time)+extractSeconds(time)+extractMilliSeconds(time);
	}
	private long extractHours(String time){
		
		double hours = 0;
		try{
			int h = -1;
			if((h = time.indexOf("h")) != -1)
				hours = Double.parseDouble(time.substring(0, h));
		} catch (NumberFormatException ex){
			System.out.println("H Exception");
			return -1;
		}		
		return (new Double(hours*60*60*1000)).longValue();				
	}
	
	private long extractMinutes(String time){
		
		double minutes = 0;
		try{
			int h = -1;
			int m = -1;
			if(time.indexOf("m") != -1)
				if(time.indexOf("ms") != time.indexOf("m")){
					m = time.indexOf("m");
					h = time.indexOf("h");
					minutes = Double.parseDouble(time.substring(h+1, m));
				}
		} catch (NumberFormatException ex){
			System.out.println("M Exception");
			return -1;
		}		
		return (new Double(minutes*60*1000)).longValue();
	}
	
	private long extractSeconds(String time){

		double seconds = 0;
		try{
			if(time.indexOf("s") != -1)
				if(time.indexOf("ms") != time.indexOf("s")-1){
					int s = time.indexOf("s");
					int m = time.indexOf("ms") != time.indexOf("m") ? time.indexOf("m") : -1;
					int h = time.indexOf("h");
					if(m > h)
						seconds = Double.parseDouble(time.substring(m+1, s));
					else
						seconds = Double.parseDouble(time.substring(h+1, s));
				}
		} catch (NumberFormatException ex){
			System.out.println("S Exception");
			return -1;
		}		
		return (new Double(seconds*1000)).longValue();
	}

	private long extractMilliSeconds(String time){
		
		double milliseconds = 0;
		try{
			if(time.indexOf("ms") != -1){
				int ms = time.indexOf("ms");
				int s = ms != time.indexOf("s")-1 ? time.indexOf("s") : -1;
				int m = ms != time.indexOf("m") ? time.indexOf("m") : -1;
				int h = time.indexOf("h");					
				
				int max = Math.max(h,m);
				max = Math.max(max,s);
				
				milliseconds = Double.parseDouble(time.substring(max+1, ms));
			}
		} catch (NumberFormatException ex){
			System.out.println("MS Exception");
			return -1;
		}		
		return (new Double(milliseconds)).longValue();
	}
	
    public void endElement(String namespaceURI,
                           String sName, // simple name
                           String qName  // qualified name
                          ) throws SAXException {
    }

    public void characters(char buf[], int offset, int len) throws SAXException {
        nl(); emit("CHARS:   ");
        String s = new String(buf, offset, len);
        if (!s.trim().equals("")) emit(s);
    }

    public void ignorableWhitespace(char buf[], int offset, int len) throws SAXException {
        // Ignore it
    }

    public void processingInstruction(String target, String data) throws SAXException {
        nl();
        emit("PROCESS: ");
        emit("<?"+target+" "+data+"?>");
    }

    //===========================================================
    // SAX ErrorHandler methods
    //===========================================================

    // treat validation errors as fatal
    public void error(SAXParseException e) throws SAXParseException {
        throw e;
    }

    // dump warnings too
    public void warning(SAXParseException err) throws SAXParseException {
        System.out.println("** Warning"
            + ", line " + err.getLineNumber()
            + ", uri " + err.getSystemId());
        System.out.println("   " + err.getMessage());
    }

    //===========================================================
    // Utility Methods ...
    //===========================================================

    // Wrap I/O exceptions in SAX exceptions, to
    // suit handler signature requirements
    private void emit(String s) throws SAXException {
        try {
            out.write(s);
            out.flush();
        } catch (IOException e) {
            throw new SAXException("I/O error", e);
        }
    }

    // Start a new line
    // and indent the next line appropriately
    private void nl() throws SAXException {
        String lineEnd =  System.getProperty("line.separator");
        try {
            out.write(lineEnd);
            //for (int i=0; i < indentLevel; i++) out.write(indentString);
        } catch (IOException e) {
            throw new SAXException("I/O error", e);
        }
    }
}
